;;; change-modes.el --- Facilities to change file modes.

;; Copyright (C) 2005 Free Software Foundation

;; Author: Michael Cadilhac <michael.cadilhac@lrde.org>
;; Keywords: file modes
;; Version: 0.1

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Commentary:

;; This package implements functions related to file modes
;; modification and symbolic notation modes parsing.

;; To change the file modes of your files, put this in your .emacs:
;; (require 'change-modes)
;; Then use M-x set-current-file-modes RET to change the current file
;; modes or M-x chmod RET to change any file modes.

;; You'll be asked the file modes you want to set, you can either
;; enter it in octal notation (644 for read/write for user, and just
;; read for group and others), or in symbolic notation in which 644 is
;; equivalent to "a=r,u+w".

;; For more information about these notations, see (info
;; "(coreutils)File permissions").


;;; Code:

;; Variables

(defvar file-modes 0
  "File modes of the file visited in the current buffer. Do not
rely on this variable to know current file modes, use `buffer-file-modes'
instead.")

(defvar file-modes-modified nil
  "Inform that the file modes has to be modified at the next save.")

(make-variable-buffer-local 'file-modes)
(make-variable-buffer-local 'file-modes-modified)


;; Hooks

;; Add the file modes setting when file is saved.
(unless (memq 'do-set-current-file-modes after-save-hook)
  (add-hook 'after-save-hook 'do-set-current-file-modes))

(defun do-set-current-file-modes ()
  "Function called after file saving to set file modes."
  (when file-modes-modified
    (set-file-modes buffer-file-name file-modes)
    (setq file-modes-modified nil)))


;; Specialized functions

(defun char-to-who (c)
  "Convert a char to a who-mask for symbolic modes notation.
C is in [ugoa] and represents users on which rights will apply."
  (cond ((= c ?u) #o4700)
	((= c ?g) #o2070)
	((= c ?o) #o1007)
	((= c ?a) #o7777)
	(t (error "%c: bad `who' character" c))))

(defun char-to-right (c &optional from)
  "Convert a char to a right-mask for symbolic modes notation.
C is in [rwxXstugo] and represents a right.
FROM (0 if non set) is the current numeric mode to process
relative rights [Xugo]."
  (unless from
    (setq from 0))
  (cond ((= c ?r) #o0444)
	((= c ?w) #o0222)
	((= c ?x) #o0111)
	((= c ?s) #o1000)
	((= c ?t) #o6000)
	;; Original file modes relative rights:
	((= c ?X) (if (= (logand from #o111) 0) 0 #o0111))
	((= c ?u) (let ((uright (logand #o4700 from)))
		    (+ uright (/ uright #o10) (/ uright #o100))))
	((= c ?g) (let ((gright (logand #o2070 from)))
		    (+ gright (/ gright #o10) (* gright #o10))))
	((= c ?o) (let ((oright (logand #o1007 from)))
		    (+ oright (* oright #o10) (* oright #o100))))
	(t (error "%c: bad right character" c))))


(defun rights-str-to-number (rights who-mask &optional from)
  "Convert rights string to rights-mask for symbolic modes notation.
RIGHTS is the rights string, it should match \"([+=-][rwxXstugo]+)+\".
WHO-MASK is the mask returned by `char-to-who'.
FROM (0 if non set) is the current numeric modes considered."
  (save-match-data
    (let ((case-fold-search nil)
	  (num-rights (if from from 0))
	  op)
      (while (memq (setq op (string-to-char rights)) '(?+ ?- ?=))
	(if (string-match "^.\\([rwxXstugo]+\\)" rights)
	    (let* ((right (string-to-list (substring rights (match-beginning 1) (match-end 1))))
		   (num-right (apply 'logior 0 (mapcar (lambda (c) (char-to-right c num-rights)) right))))
	      (setq num-right (logand who-mask num-right)
		    num-rights (cond ((= op ?+) (logior num-rights num-right))
				     ((= op ?-) (logand num-rights (lognot num-right)))
				     (t (logior (logand num-rights (lognot who-mask)) num-right)))
		    rights (substring rights (match-end 1))))
	  (error "Parse error in rights string near `%s'" (substring rights 0))))
      num-rights)))

(defun symbolic-modes-to-number (modes &optional from)
  "Convert file modes denoted by the symbolic notation to numeric modes.
MODES is the string to convert, it should match
\"[ugoa]*([+-=][rwxXstugo]+)+,...\". Look at
`(info \"(coreutils)File permissions\")' for more informations about this
notation.
FROM (0 if non set) is the current numeric modes considered."
  (let ((case-fold-search nil)
	(num-modes (if from from 0)))
    (while (not (string= modes ""))
      (if (string-match "^\\([ugoa]*\\)\\([+=-][rwxXstugo]+\\)+\\(,\\|\\)" modes)
	  (let* ((who (string-to-list (substring modes (match-beginning 1) (match-end 1))))
		 (num-who (apply 'logior 0 (mapcar 'char-to-who who))))
	    (when (= num-who 0)
	      (setq num-who (default-file-modes)))
	    (setq num-modes
		  (rights-str-to-number (substring modes (match-end 1))
					num-who num-modes)
		  modes (substring modes (match-end 3))))
	(error "Parse error in modes near %s" (substring modes 0))))
    num-modes))


;; Main functions

;;;###autoload
(defun file-modes-to-number (modes &optional from)
  "Convert file modes denoted by MODES to numeric modes.
MODES is either a string representing the modes in octal or in symbolic
notation, or a number.
FROM (0 if non set) is the numeric modes to which modes will refer is
the symbolic notation is used."
  (if (stringp modes)
      (if (string-match "^[0-7]+$" modes)
	  (string-to-number modes 8)
	(symbolic-modes-to-number modes from))
    modes))

;;;###autoload
(defun buffer-file-modes (&optional buffer)
  "Return the file modes of the file currently visited by BUFFER.
If BUFFER doesn't visit a saved file, return the default file modes.
If BUFFER is not set, current buffer is used."
  (with-current-buffer (if buffer buffer (current-buffer))
    (if (and buffer-file-name (file-exists-p buffer-file-name))
	(file-modes buffer-file-name)
      (default-file-modes))))

;;;###autoload
(defun set-current-file-modes (modes)
  "Set the file modes of the current file. It will be modified on saving.
MODES is the file mode in symbolic or octal notation, or a number."
  (interactive (list (read-string "File modes (symbolic or octal): ")))
  (unless (and (stringp modes) (string= modes ""))
    (setq file-modes (file-modes-to-number modes (buffer-file-modes))
	  file-modes-modified t)
    (set-buffer-modified-p t)))

;;;###autoload
(defun chmod (filename modes)
  "Set MODES to be the file modes of FILENAME.
MODES is the file mode in symbolic or octal notation, or a number."
  (interactive (list (read-file-name "Change mode: " nil nil t)
		     (read-string "File modes (symbolic or octal): ")))
  (unless (file-exists-p filename)
    (error "File %s doesn't exist" filename))
  (unless (and (stringp modes) (string= modes ""))
    (set-file-modes filename (file-modes-to-number modes (file-modes filename)))))

(provide 'change-modes)

;;; change-modes.el ends here
